"""
This module defines an :py:class:`aiohttp.ClientSession` adapter that
returns :py:class:`twisted.internet.defer.Deferred` responses.
"""

# Third party imports
try:
    from twisted.internet import threads
except ImportError:
    threads = None

# Local imports
from uplink.clients import helpers, interfaces, register


class TwistedClient(interfaces.HttpClientAdapter):
    """
    Client that returns :py:class:`twisted.internet.defer.Deferred`
    responses.

    Note:
        This client is an optional feature and requires the :py:mod:`twisted`
        package. For example, here's how to install this extra using pip::

            $ pip install uplink[twisted]

    Args:
        session (:py:class:`requests.Session`, optional): The session
            that should handle sending requests. If this argument is
            omitted or set to :py:obj:`None`, a new session will be
            created.
    """

    def __init__(self, session=None):
        if threads is None:
            raise NotImplementedError("twisted is not installed.")
        self._requests = register.get_client(session)

    def create_request(self):
        return Request(self._requests.create_request())


class Request(helpers.ExceptionHandlerMixin, interfaces.Request):

    def __init__(self, proxy):
        self._proxy = proxy
        self._callback = None

    def send(self, method, url, extras):
        deferred = threads.deferToThread(
            self._proxy.send,
            method,
            url,
            extras
        )
        if self._callback is not None:
            deferred.addCallback(self._callback)
        deferred.addErrback(self.handle_failure)
        return deferred

    def add_callback(self, callback):
        self._callback = callback

    def handle_failure(self, failure):
        tb = failure.getTracebackObject()
        self._exception_handler.handle(failure.type, failure.value, tb)
        failure.raiseException()



"""This module implements the auth layer."""

# Standard library imports
import collections

# Third-party imports
from requests import auth

# Local imports
from uplink import utils

__all__ = []


def get_auth(auth_object=None):
    if auth_object is None:
        return utils.no_op
    elif isinstance(auth_object, collections.Iterable):
        return BasicAuth(*auth_object)
    elif callable(auth_object):
        return auth_object
    else:
        raise ValueError(
            "Invalid authentication strategy: %s" % auth_object
        )


class BasicAuth(object):
    """Authorizes requests using HTTP Basic Authentication."""

    def __init__(self, username, password):
        self._username = username
        self._password = password

    @property
    def _auth_str(self):
        return auth._basic_auth_str(
            self._username,
            self._password
        )

    def __call__(self, request_builder):
        request_builder.info["headers"]["Authorization"] = self._auth_str


class ProxyAuth(BasicAuth):

    def __call__(self, request_builder):
        request_builder.info["headers"]["Proxy-Authorization"] = self._auth_str



class Converter(object):

    def convert(self, value):
        raise NotImplementedError


class ConverterFactory(object):

    def make_response_body_converter(self, type_, argument_annotations,
                                     method_annotations):
        raise NotImplementedError

    def make_request_body_converter(self, type_, argument_annotations,
                                    method_annotations):
        raise NotImplementedError

    def make_string_converter(self, type_, argument_annotations,
                              method_annotations):
        raise NotImplementedError



# Standard library imports
import inspect

# Local imports
from uplink import exceptions, helpers, hooks, interfaces, types


__all__ = [
    "headers",
    "form_url_encoded",
    "multipart",
    "json",
    "timeout",
    "returns",
    "args",
    "response_handler",
    "error_handler",
    "inject",
]


class HttpMethodNotSupport(exceptions.AnnotationError):
    message = (
        "Error on method `%s`: annotation %s is not supported with %s "
        "commands."
    )

    def __init__(self, request_definition_builder, annotation):
        self.message = self.message % (
            request_definition_builder.__name__,
            type(annotation),
            request_definition_builder.method.upper()
        )


class MethodAnnotationHandlerBuilder(interfaces.AnnotationHandlerBuilder):

    def __init__(self):
        self._class_annotations = list()
        self._method_annotations = list()

    def add_annotation(self, annotation, *args_, **kwargs):
        if kwargs.get("is_class", False):
            self._class_annotations.append(annotation)
        else:
            self._method_annotations.append(annotation)
        super(MethodAnnotationHandlerBuilder, self).add_annotation(annotation)
        return annotation

    def build(self):
        return MethodAnnotationHandler(
            self._class_annotations + self._method_annotations
        )


class MethodAnnotationHandler(interfaces.AnnotationHandler):

    def __init__(self, method_annotations):
        self._method_annotations = list(method_annotations)

    @property
    def annotations(self):
        return iter(self._method_annotations)

    def handle_builder(self, request_builder):
        for annotation in self._method_annotations:
            annotation.modify_request(request_builder)


class MethodAnnotation(interfaces.Annotation):
    http_method_whitelist = None

    @classmethod
    def _is_static_call(cls, *args_, **kwargs):
        if super(MethodAnnotation, cls)._is_static_call(*args_, **kwargs):
            return True
        try:
            is_class = inspect.isclass(args_[0])
        except IndexError:
            return False
        else:
            return is_class and not (kwargs or args_[1:])

    def __call__(self, class_or_builder):
        if inspect.isclass(class_or_builder):
            builders = helpers.get_api_definitions(class_or_builder)

            for name, builder in builders:
                builder.method_handler_builder.add_annotation(
                    self,
                    is_class=True,
                )
                helpers.set_api_definition(class_or_builder, name, builder)
        else:
            class_or_builder.method_handler_builder.add_annotation(self)
        return class_or_builder

    def modify_request_definition(self, request_definition_builder):
        if self.http_method_whitelist is not None:
            method = request_definition_builder.method.upper()
            if method not in self.http_method_whitelist:
                raise HttpMethodNotSupport(request_definition_builder, self)

    def modify_request(self, request_builder):
        pass


# noinspection PyPep8Naming
class headers(MethodAnnotation):
    """
    A decorator that adds static headers for API calls.

    .. code-block:: python

        @headers({"User-Agent": "Uplink-Sample-App})
        @get("/user")
        def get_user(self):
            \"""Get the current user\"""

    When used as a class decorator, :py:class:`headers` applies to
    all consumer methods bound to the class:

    .. code-block:: python

        @headers({"Accept": "application/vnd.github.v3.full+json")
        class GitHub(Consumer):
            ...

    :py:class:`headers` takes the same arguments as :py:class:`dict`.

    Args:
        *arg: A dict containing header values.
        **kwargs: More header values.
    """

    def __init__(self, arg, **kwargs):
        if isinstance(arg, str):
            key, value = self._get_header(arg)
            self._headers = {key: value}
        elif isinstance(arg, list):
            self._headers = dict(self._get_header(a) for a in arg)
        else:
            self._headers = dict(arg, **kwargs)

    def _get_header(self, arg):
        return map(str.strip, arg.split(":"))

    def modify_request(self, request_builder):
        """Updates header contents."""
        request_builder.info["headers"].update(self._headers)


# noinspection PyPep8Naming
class form_url_encoded(MethodAnnotation):
    """
    URL-encodes the request body.

    Used on POST/PUT/PATCH request. It url-encodes the body of the
    message and sets the appropriate ``Content-Type`` header. Further,
    each field argument should be annotated with
    :py:class:`uplink.Field`.

    Example:
        .. code-block:: python

            @form_url_encoded
            @post("/users/edit")
            def update_user(self, first_name: Field, last_name: Field):
                \"""Update the current user.\"""
    """
    _can_be_static = True

    # XXX: Let `requests` handle building urlencoded syntax.
    # def modify_request(self, request_builder):
    #     request_builder.info.headers(
    #         {"Content-Type": "application/x-www-form-urlencoded"}
    #     )


# noinspection PyPep8Naming
class multipart(MethodAnnotation):
    """
    Sends multipart form data.

    Multipart requests are commonly used to upload files to a server.
    Further, annotate each part argument with :py:class:`Part`.

    Example:

        .. code-block:: python

            @multipart
            @put(/user/photo")
            def update_user(self, photo: Part, description: Part):
                \"""Upload a user profile photo.\"""
    """
    _can_be_static = True

    # XXX: Let `requests` handle building multipart syntax.
    # def modify_request(self, request_builder):
    #     request_builder.info.headers(
    #         {"Content-Type": "multipart/form-data"}
    #     )


# noinspection PyPep8Naming
class json(MethodAnnotation):
    """Use as a decorator to make JSON requests.

    You should annotate a method argument with `uplink.Body` which
    indicates that the argument's value should become the request's
    body. :py:class:`uplink.Body` has to be either a dict or a subclass
    of py:class:`collections.Mapping`.

    Example:
        .. code-block:: python

            @json
            @patch(/user")
            def update_user(self, **info: Body):
                \"""Update the current user.\"""
    """
    _can_be_static = True

    def modify_request(self, request_builder):
        """Modifies JSON request."""
        try:
            request_builder.info["json"] = request_builder.info.pop("data")
        except KeyError:
            pass


# noinspection PyPep8Naming
class timeout(MethodAnnotation):
    """
    Time to wait for a server response before giving up.

    When used on other decorators it specifies how long (in secs) a
    decorator should wait before giving up.

    Example:
        .. code-block:: python

            @timeout(60)
            @get("/user/posts")
            def get_posts(self):
                \"""Fetch all posts for the current users.\"""

    When used as a class decorator, :py:class:`timeout` applies to all
    consumer methods bound to the class.

    Args:
        seconds (int): An integer used to indicate how long should the
            request wait.
    """
    def __init__(self, seconds):
        self._seconds = seconds

    def modify_request(self, request_builder):
        """Modifies request timeout."""
        request_builder.info["timeout"] = self._seconds


# noinspection PyPep8Naming
class returns(MethodAnnotation):
    """
    Specify the function's return annotation for Python 2.7
    compatibility.

    In Python 3, to provide a consumer method's return type, you can
    set the it as the method's return annotation:

    .. code-block:: python

        @get("/users/{username}")
        def get_user(self, username) -> UserSchema:
            \"""Get a specific user.\"""

    For Python 2.7 compatibility, you can use this decorator instead:

    .. code-block:: python

        @returns(UserSchema)
        @get("/users/{username}")
        def get_user(self, username):
            \"""Get a specific user.\"""
    """

    def __init__(self, type):
        self._type = type

    def modify_request(self, request_builder):
        request_builder.return_type = self._type


# noinspection PyPep8Naming
class args(MethodAnnotation):
    """
    Annotate method arguments for Python 2.7 compatibility.

    Arrange annotations in the same order as their corresponding
    function arguments.

    Example:
        .. code-block:: python

            @args(Path, Query)
            @get("/users/{username})
            def get_user(self, username, visibility):
                \"""Get a specific user.\"""

    Use keyword args to target specific method parameters.

    Example:
        .. code-block:: python

            @args(visibility=Query)
            @get("/users/{username})
            def get_user(self, username, visibility):
                \"""Get a specific user.\"""

    Args:
        *annotations: Any number of annotations.
        **more_annotations: More annotations, targeting specific method
           arguments.
    """
    def __init__(self, *annotations, **more_annotations):
        self._annotations = annotations
        self._more_annotations = more_annotations

    def __call__(self, obj):
        if inspect.isfunction(obj):
            handler = types.ArgumentAnnotationHandlerBuilder.from_func(obj)
            self._helper(handler)
            return obj
        else:
            return super(args, self).__call__(obj)

    def _helper(self, builder):
        builder.set_annotations(self._annotations, **self._more_annotations)

    def modify_request_definition(self, request_definition_builder):
        """Modifies dynamic requests with given annotations"""
        self._helper(request_definition_builder.argument_handler_builder)


class _InjectableMethodAnnotation(MethodAnnotation):
    def modify_request(self, request_builder):
        request_builder.add_transaction_hook(self)


# noinspection PyPep8Naming
class response_handler(_InjectableMethodAnnotation, hooks.ResponseHandler):
    """
    A decorator for creating custom response handlers.

    To register a function as a custom response handler, decorate the
    function with this class. The decorated function should accept a single
    positional argument, an HTTP response object:

    Example:
        .. code-block:: python

            @response_handler
            def raise_for_status(response):
                response.raise_for_status()
                return response

    Then, to apply custom response handling to a request method, simply
    decorate the method with the registered response handler:

    Example:
        .. code-block:: python

            @raise_for_status
            @get("/user/posts")
            def get_posts(self):
                \"""Fetch all posts for the current users.\"""

    To apply custom response handling on all request methods of a
    :py:class:`uplink.Consumer` subclass, simply decorate the class with
    the registered response handler:

    Example:
        .. code-block:: python

            @raise_for_status
            class GitHub(Consumer):
               ...

    .. versionadded:: 0.4.0
    """


# noinspection PyPep8Naming
class error_handler(_InjectableMethodAnnotation, hooks.ExceptionHandler):
    """
    A decorator for creating custom error handlers.

    To register a function as a custom error handler, decorate the
    function with this class. The decorated function should accept three
    positional arguments: (1) the type of the exception, (2) the
    exception instance raised, and (3) a traceback instance.

    Example:
        .. code-block:: python

            @error_handler
            def raise_api_error(exc_type, exc_val, exc_tb):
                # wrap client error with custom API error
                ...

    Then, to apply custom error handling to a request method, simply
    decorate the method with the registered error handler:

    Example:
        .. code-block:: python

            @raise_api_error
            @get("/user/posts")
            def get_posts(self):
                \"""Fetch all posts for the current users.\"""

    To apply custom error handling on all request methods of a
    :py:class:`uplink.Consumer` subclass, simply decorate the class with
    the registered error handler:

    Example:
        .. code-block:: python

            @raise_api_error
            class GitHub(Consumer):
               ...

    .. versionadded:: 0.4.0

    Note:
        Error handlers can not completely suppress exceptions. The
        original exception is thrown if the error handler doesn't throw
        anything.
    """


# noinspection PyPep8Naming
class inject(_InjectableMethodAnnotation, hooks.TransactionHookChain):
    """
    A decorator that applies one or more hooks to a request method.

    Example:
        .. code-block:: python

            @inject(Query("sort").with_value("pushed"))
            @get("users/{user}/repos")
            def list_repos(self, user):
                \"""Lists user's public repos by latest pushed.\"""

    .. versionadded:: 0.4.0
    """



# Standard library imports
import functools

# Local imports
from uplink import converters, decorators, exceptions, interfaces, types, utils

__all__ = ["get", "head", "put", "post", "patch", "delete"]


class MissingArgumentAnnotations(exceptions.InvalidRequestDefinition):
    message = "Missing annotation for argument(s): '%s'."
    implicit_message = " (Implicit path variables: '%s')"

    def __init__(self, missing, path_variables):
        missing, path_variables = list(missing), list(path_variables)
        self.message = self.message % "', '".join(missing)
        if path_variables:
            self.message += self.implicit_message % "', '".join(path_variables)


class MissingUriVariables(exceptions.InvalidRequestDefinition):
    message = "On uri template %s, some variables are not handled: %s"

    def __init__(self, uri, remaining_variables):
        self.message = self.message % (uri, "', '".join(remaining_variables))


class HttpMethodFactory(object):

    def __init__(self, method):
        self._method = method

    def __call__(self, uri=None):
        if callable(uri):
            return HttpMethod(self._method)(uri)
        else:
            return HttpMethod(self._method, uri)


class HttpMethod(object):

    def __init__(self, method, uri=None):
        self._method = method
        self._uri = uri

    def __call__(self, func):
        spec = utils.get_arg_spec(func)
        arg_handler = types.ArgumentAnnotationHandlerBuilder(func, spec.args)
        builder = RequestDefinitionBuilder(
            self._method,
            URIDefinitionBuilder(self._uri),
            arg_handler,
            decorators.MethodAnnotationHandlerBuilder()
        )
        arg_handler.add_annotations_from_spec(spec)
        if spec.return_annotation is not None:
            builder = decorators.returns(spec.return_annotation)(builder)
        functools.update_wrapper(builder, func)
        return builder


class URIDefinitionBuilder(interfaces.UriDefinitionBuilder):

    def __init__(self, uri):
        self._uri = uri
        self._is_dynamic = False
        self._uri_variables = set()

    @property
    def is_static(self):
        return self._uri is not None

    @property
    def is_dynamic(self):
        return self._is_dynamic

    @is_dynamic.setter
    def is_dynamic(self, is_dynamic):
        if self.is_static:
            raise ValueError(
                "Failed to set dynamic URI as URI is already defined: %s"
                % self._uri
            )
        self._is_dynamic = is_dynamic

    def add_variable(self, name):
        if self.is_static and name not in self.remaining_variables:
            raise ValueError(
                "`%s` is not a variable on the uri %s" % (name, self._uri)
            )
        self._uri_variables.add(name)

    @property
    def remaining_variables(self):
        return utils.URIBuilder.variables(self._uri) - self._uri_variables

    def build(self):
        if self.remaining_variables:
            raise MissingUriVariables(self._uri, self.remaining_variables)
        return self._uri


class RequestDefinitionBuilder(interfaces.RequestDefinitionBuilder):

    def __init__(self, method, uri, argument_handler_builder,
                 method_handler_builder):
        self._method = method
        self._uri = uri
        self._argument_handler_builder = argument_handler_builder
        self._method_handler_builder = method_handler_builder

        self._argument_handler_builder.listener = self._notify
        self._method_handler_builder.listener = self._notify

    def _notify(self, annotation):
        annotation.modify_request_definition(self)

    @property
    def method(self):
        return self._method

    @property
    def uri(self):
        return self._uri

    @property
    def argument_handler_builder(self):
        return self._argument_handler_builder

    @property
    def method_handler_builder(self):
        return self._method_handler_builder

    def _auto_fill_remaining_arguments(self):
        uri_vars = set(self.uri.remaining_variables)
        missing = list(self.argument_handler_builder.missing_arguments)
        still_missing = set(missing) - uri_vars

        # Preserve order of function parameters.
        matching = [p for p in missing if p in uri_vars]

        if still_missing:
            raise MissingArgumentAnnotations(still_missing, matching)

        path_vars = dict.fromkeys(matching, types.Path)
        self.argument_handler_builder.set_annotations(path_vars)

    def build(self):
        if not self._argument_handler_builder.is_done():
            self._auto_fill_remaining_arguments()
        argument_handler = self._argument_handler_builder.build()
        method_handler = self._method_handler_builder.build()
        uri = self._uri.build()
        return RequestDefinition(
            self._method,
            uri,
            argument_handler,
            method_handler
        )


class RequestDefinition(interfaces.RequestDefinition):

    def __init__(self, method, uri, argument_handler, method_handler):
        self._method = method
        self._uri = uri
        self._argument_handler = argument_handler
        self._method_handler = method_handler

    @property
    def argument_annotations(self):
        return iter(self._argument_handler.annotations)

    @property
    def method_annotations(self):
        return iter(self._method_handler.annotations)

    def make_converter_registry(self, converters_):
        return converters.ConverterFactoryRegistry(
            converters_,
            argument_annotations=self.argument_annotations,
            request_annotations=self.method_annotations
        )

    def define_request(self, request_builder, func_args, func_kwargs):
        request_builder.method = self._method
        request_builder.url = utils.URIBuilder(self._uri)
        self._argument_handler.handle_call(
            request_builder, func_args, func_kwargs)
        self._method_handler.handle_builder(request_builder)
        request_builder.url = request_builder.url.build()


get = HttpMethodFactory("GET").__call__
head = HttpMethodFactory("HEAD").__call__
put = HttpMethodFactory("PUT").__call__
post = HttpMethodFactory("POST").__call__
patch = HttpMethodFactory("PATCH").__call__
delete = HttpMethodFactory("DELETE").__call__



# Local imports
from uplink.__about__ import __version__
from uplink.clients import *
from uplink.converters import *
from uplink.builder import *
from uplink.commands import *
from uplink.exceptions import *
from uplink.decorators import *
from uplink.types import *



"""
This module defines a converter that uses :py:mod:`marshmallow` schemas
to deserialize and serialize values.
"""

# Standard library imports
import inspect

# Local imports
from uplink.converters import interfaces


class MarshmallowConverter(interfaces.ConverterFactory):
    """
    A converter that serializes and deserializes values using
    :py:mod:`marshmallow` schemas.

    To deserialize JSON responses into Python objects with this
    converter, define a :py:class:`marshmallow.Schema` subclass and set
    it as the return annotation of a consumer method:

    .. code-block:: python

        @get("/users")
        def get_users(self, username) -> UserSchema():
            '''Fetch a single user'''

    Also, when instantiating a consumer, be sure to set this class as
    a converter for the instance:

    .. code-block:: python

        github = GitHub(BASE_URL, converter=MarshmallowConverter())

    Note:

        This converter is an optional feature and requires the :py:mod:`marshmallow`
        package. For example, here's how to install this feature using pip::

            $ pip install uplink[marshmallow]
    """
    try:
        import marshmallow
    except ImportError:  # pragma: no cover
        marshmallow = None

    def __init__(self):
        if self.marshmallow is None:
            raise ImportError("No module named 'marshmallow'")

    class ResponseBodyConverter(interfaces.Converter):

        def __init__(self, schema):
            self._schema = schema

        def convert(self, response):
            try:
                json = response.json()
            except AttributeError:
                return response
            else:
                return self._schema.load(json).data

    class RequestBodyConverter(interfaces.Converter):
        def __init__(self, schema):
            self._schema = schema

        def convert(self, value):
            return self._schema.dump(value).data

    @classmethod
    def _get_schema(cls, type_):
        if inspect.isclass(type_) and issubclass(type_, cls.marshmallow.Schema):
            return type_()
        elif isinstance(type_, cls.marshmallow.Schema):
            return type_
        raise ValueError("Expected marshmallow.Scheme subclass or instance.")

    def _make_converter(self, converter_cls, type_):
        try:
            # Try to generate schema instance from the given type.
            schema = self._get_schema(type_)
        except ValueError:
            # Failure: the given type is not a `marshmallow.Schema`.
            return None
        else:
            return converter_cls(schema)

    def make_request_body_converter(self, type_, *args, **kwargs):
        """
        Constructs a :py:class:`uplink.converters.interfaces.Converter`
        subclass that serializes values using a
        :py:class:`marshmallow.Schema`.
        """
        return self._make_converter(self.RequestBodyConverter, type_)

    def make_response_body_converter(self, type_, *args, **kwargs):
        """
        Constructs a :py:class:`uplink.converters.interfaces.Converter`
        subclass that deserializes values using a
        :py:class:`marshmallow.Schema`.
        """
        return self._make_converter(self.ResponseBodyConverter, type_)

    def make_string_converter(self, type_, *args, **kwargs):
        return None



"""
This module provides a class for defining custom handling for specific
points of an HTTP transaction.
"""

__all__ = [
    "TransactionHook",
    "RequestAuditor",
    "ResponseHandler"
]


class TransactionHook(object):
    """
    A utility class providing methods that define hooks for specific
    points of an HTTP transaction.
    """

    def audit_request(self, request_builder):  # pragma: no cover
        """Inspects details of a request before it is sent."""
        pass

    def handle_response(self, response):
        """
        Handles a response object from the server.

        Args:
            response: The received HTTP response.
        """
        return response

    def handle_exception(self, exc_type, exc_val, exc_tb):  # pragma: no cover
        """
        Handles an exception thrown while waiting for a response from
        the server.

        Args:
            exc_type: The type of the exception.
            exc_val: The exception instance raised.
            exc_tb: A traceback instance.
        """
        pass


class TransactionHookChain(TransactionHook):
    """
    A chain that conjoins several transaction hooks into a single
    object.

    A method call on this composite object invokes the corresponding
    method on all hooks in the chain.
    """

    def __init__(self, hook, *hooks):
        self._hooks = (hook,)
        if not hooks:
            self.handle_response = hook.handle_response
        else:
            self._hooks += hooks

    def audit_request(self, *args, **kwargs):
        for hook in self._hooks:
            hook.audit_request(*args, **kwargs)

    def handle_response(self, response, *args, **kwargs):
        for hook in self._hooks:
            response = hook.handle_response(response, *args, **kwargs)
        return response

    def handle_exception(self, *args, **kwargs):
        for hook in self._hooks:
            hook.handle_exception(*args, **kwargs)


class RequestAuditor(TransactionHook):
    """
    Transaction hook that inspects requests using a function provided at
    time of instantiation.
    """

    def __init__(self, auditor):
        self.audit_request = auditor


class ResponseHandler(TransactionHook):
    """
    Transaction hook that handles responses using a function provided at
    time of instantiation.
    """

    def __init__(self, handler):
        self.handle_response = handler


class ExceptionHandler(TransactionHook):
    """
    Transaction hook that handles an exception thrown while waiting for
    a response, using the provided function.
    """

    def __init__(self, exception_handler):
        self.handle_exception = exception_handler




class HttpClientAdapter(object):
    """An adapter of an HTTP client library."""

    def create_request(self):
        raise NotImplementedError


class Request(object):

    def send(self, method, url, extras):
        raise NotImplementedError

    def add_callback(self, callback):
        raise NotImplementedError

    def add_exception_handler(self, exception_handler):
        raise NotImplementedError



# Standard library imports
import collections
import functools
import warnings

# Local imports
from uplink import (
    auth as auth_,
    clients,
    converters,
    exceptions,
    helpers,
    hooks,
    interfaces,
    utils,
    types
)
from uplink.converters import keys

__all__ = ["build", "Consumer"]


class RequestPreparer(object):

    def __init__(self, builder):
        self._hooks = list(builder.hooks)
        self._client = builder.client
        self._base_url = str(builder.base_url)
        self._converters = list(builder.converters)
        self._auth = builder.auth

    def _join_url_with_base(self, url):
        return utils.urlparse.urljoin(self._base_url, url)

    def _get_hook_chain(self, contract):
        chain = list(contract.transaction_hooks)
        converter = contract.get_converter(
            keys.CONVERT_FROM_RESPONSE_BODY,
            contract.return_type)
        if converter is not None:
            # Found a converter that can handle the return type.
            chain.append(hooks.ResponseHandler(converter.convert))
        chain.extend(self._hooks)
        return chain

    @staticmethod
    def apply_hooks(chain, request_builder, sender):
        hook = hooks.TransactionHookChain(*chain)
        hook.audit_request(request_builder)
        sender.add_callback(hook.handle_response)
        sender.add_exception_handler(hook.handle_exception)

    def prepare_request(self, request_builder):
        request_builder.url = self._join_url_with_base(request_builder.url)
        self._auth(request_builder)
        sender = self._client.create_request()
        chain = self._get_hook_chain(request_builder)
        if chain:
            self.apply_hooks(chain, request_builder, sender)
        return sender.send(
            request_builder.method,
            request_builder.url,
            request_builder.info
        )

    def create_request_builder(self, definition):
        registry = definition.make_converter_registry(self._converters)
        return helpers.RequestBuilder(registry)


class CallFactory(object):
    def __init__(self, request_preparer, request_definition):
        self._request_preparer = request_preparer
        self._request_definition = request_definition

    def __call__(self, *args, **kwargs):
        builder = self._request_preparer.create_request_builder(
            self._request_definition)
        self._request_definition.define_request(builder, args, kwargs)
        return self._request_preparer.prepare_request(builder)


class Builder(interfaces.CallBuilder):
    """The default callable builder."""

    def __init__(self):
        self._base_url = ""
        self._hooks = []
        self._client = clients.get_client()
        self._converters = collections.deque()
        self._converters.append(converters.StandardConverter())
        self._auth = auth_.get_auth()

    @property
    def client(self):
        return self._client

    @client.setter
    def client(self, client):
        if client is not None:
            self._client = clients.get_client(client)

    @property
    def hooks(self):
        return iter(self._hooks)

    def add_hook(self, *hooks_):
        self._hooks.extend(hooks_)

    @property
    def base_url(self):
        return self._base_url

    @base_url.setter
    def base_url(self, base_url):
        self._base_url = base_url

    @property
    def converters(self):
        return iter(self._converters)

    def add_converter(self, *converters_):
        self._converters.extendleft(converters_)

    @property
    def auth(self):
        return self._auth

    @auth.setter
    def auth(self, auth):
        if auth is not None:
            self._auth = auth_.get_auth(auth)

    def build(self, definition):
        """
        Creates a callable that uses the provided definition to execute
        HTTP requests when invoked.
        """
        return CallFactory(RequestPreparer(self), definition)


class ConsumerMethod(object):
    """
    A wrapper around a :py:class`interfaces.RequestDefinitionBuilder`
    instance bound to a :py:class:`Consumer` subclass, mainly responsible
    for controlling access to the instance.
    """

    def __init__(self, owner_name, attr_name, request_definition_builder):
        self._request_definition_builder = request_definition_builder
        self._owner_name = owner_name
        self._attr_name = attr_name
        self._request_definition = self._build_definition()

    def _build_definition(self):
        try:
            return self._request_definition_builder.build()
        except exceptions.InvalidRequestDefinition as error:
            # TODO: Find a Python 2.7 compatible way to reraise
            raise exceptions.UplinkBuilderError(
                self._owner_name,
                self._attr_name,
                error)

    def __get__(self, instance, owner):
        if instance is None:
            return self._request_definition_builder
        else:
            return instance._builder.build(self._request_definition)


class ConsumerMeta(type):

    @staticmethod
    def _wrap_if_definition(cls_name, key, value):
        if isinstance(value, interfaces.RequestDefinitionBuilder):
            value = ConsumerMethod(cls_name, key, value)
        return value

    @staticmethod
    def _set_init_handler(namespace):
        try:
            init = namespace["__init__"]
        except KeyError:
            pass
        else:
            builder = types.ArgumentAnnotationHandlerBuilder.from_func(init)
            handler = builder.build()

            @functools.wraps(init)
            def new_init(self, *args, **kwargs):
                init(self, *args, **kwargs)
                call_args = utils.get_call_args(init, self, *args, **kwargs)
                f = functools.partial(
                    handler.handle_call_args, call_args=call_args
                )
                hook = hooks.RequestAuditor(f)
                self._builder.add_hook(hook)

            namespace["__init__"] = new_init

    def __new__(mcs, name, bases, namespace):
        mcs._set_init_handler(namespace)

        # Wrap all definition builders with a special descriptor that
        # handles attribute access behavior.
        for key, value in namespace.items():
            namespace[key] = mcs._wrap_if_definition(name, key, value)
        return super(ConsumerMeta, mcs).__new__(mcs, name, bases, namespace)

    def __setattr__(cls, key, value):
        value = cls._wrap_if_definition(cls.__name__, key, value)
        super(ConsumerMeta, cls).__setattr__(key, value)


_Consumer = ConsumerMeta("_Consumer", (), {})


class Consumer(_Consumer):

    def __init__(
            self,
            base_url="",
            client=None,
            converter=(),
            auth=None,
            hook=()
    ):
        self._builder = Builder()
        self._builder.base_url = base_url
        if isinstance(converter, converters.interfaces.ConverterFactory):
            converter = (converter,)
        self._builder.add_converter(*converter)
        if isinstance(hook, hooks.TransactionHook):
            hook = (hook,)
        self._builder.add_hook(*hook)
        self._builder.auth = auth
        self._builder.client = client

    def _inject(self, hook, *more_hooks):
        self._builder.add_hook(hook, *more_hooks)


def build(service_cls, *args, **kwargs):
    name = service_cls.__name__
    warnings.warn(
        "`uplink.build` is deprecated and will be removed in v1.0.0. "
        "To construct a consumer instance, have `{0}` inherit "
        "`uplink.Consumer` then instantiate (e.g., `{0}(...)`). ".format(name),
        DeprecationWarning
    )
    consumer = type(name, (service_cls, Consumer), dict(service_cls.__dict__))
    return consumer(*args, **kwargs)



# Standard library imports
import os
from setuptools import setup, find_packages
import sys


def read(filename):
    with open(filename) as stream:
        return stream.read()


# Read package metadata from __about__.py, to avoid importing the whole
# package prior to installation.
about = dict()
with open(os.path.join("uplink", "__about__.py")) as fp:
    exec(fp.read(), about)
    about = dict((k.strip("_"), about[k]) for k in about)

install_requires = [
    "requests>=2.18.0",
    "uritemplate>=3.0.0",
]

extras_require = {
    "aiohttp": ["aiohttp>=2.3.0"],
    "twisted": ["twisted>=17.1.0"],
    "marshmallow": ["marshmallow>=2.15.0"]
}

metadata = dict({
    "name": "uplink",
    "author": "P. Raj Kumar",
    "author_email": "raj.pritvi.kumar@gmail.com",
    "url": "https://uplink.readthedocs.io/",
    "license": "MIT",
    "description": "A Declarative HTTP Client for Python.",
    "long_description": read("README.rst"),
    "classifiers": [
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7"
    ],
    "keywords": "http api rest client retrofit",
    "packages": find_packages(exclude=("tests",)),
    "install_requires": install_requires,
    "extras_require": extras_require
}, **about)

if __name__ == "__main__":
    setup(**metadata)



# Standard library imports
import collections
import functools
import json

# Local imports
from uplink.converters import interfaces, keys
from uplink.converters.marshmallow_ import MarshmallowConverter

__all__ = ["StandardConverter", "MarshmallowConverter"]


class Cast(interfaces.Converter):
    def __init__(self, caster, converter):
        self._cast = caster
        self._converter = converter

    def convert(self, value):
        if callable(self._cast):
            value = self._cast(value)
        return self._converter.convert(value)


class RequestBodyConverter(interfaces.Converter):
    def convert(self, value):
        if isinstance(value, str):
            return value
        else:
            return json.loads(
                json.dumps(value, default=lambda obj: obj.__dict__)
            )


class StringConverter(interfaces.Converter):
    def convert(self, value):
        return str(value)


class StandardConverter(interfaces.ConverterFactory):
    """
    The default converter, this class seeks to provide sane alternatives
    for (de)serialization when all else fails -- e.g., no other
    converters could handle a particular type.
    """

    def make_response_body_converter(self, type_, *args, **kwargs):
        return None  # pragma: no cover

    def make_request_body_converter(self, type_, *args, **kwargs):
        return Cast(type_, RequestBodyConverter())  # pragma: no cover

    def make_string_converter(self, type_, *args, **kwargs):
        return Cast(type_, StringConverter())  # pragma: no cover


class ConverterFactoryRegistry(collections.Mapping):
    """
    A registry that chains together
    :py:class:`interfaces.ConverterFactory` instances.

    When queried for a factory that can handle a particular converter
    type (e.g., ``keys.CONVERT_TO_REQUEST_BODY``), the registry
    traverses the chain until it finds a converter factory that can
    handle the request (i.e., the type's associated method returns a
    value other than ``None``).

    Here's an example -- it's contrived but effectively details the
    expected pattern of usage::

        # Create a registry with a single factory in its chain.
        registry = ConverterFactoryRegistry((StandardConverter,))

        # Get a callable that returns converters for turning arbitrary
        # objects into strings.
        get_str_converter_for_type = registry[keys.CONVERT_TO_STRING]

        # Traverse the chain to find a converter that can handle
        # converting ints into strings.
        converter = get_str_converter_for_type(int)

    Args:
        factories: An iterable of converter factories. Factories that
            appear earlier in the chain are given the opportunity to
            handle a request before those that appear later.
    """

    #: A mapping of keys to callables. Each callable value accepts a
    #: single argument, a :py:class:`interfaces.ConverterFactory`
    #: subclass, and returns another callable, which should return a
    #: :py:`interfaces.Converter` instance.
    _converter_factory_registry = {}

    def __init__(self, factories, *args, **kwargs):
        self._factories = tuple(factories)
        self._args = args
        self._kwargs = kwargs

    @property
    def factories(self):
        """
        Yields the registry's chain of converter factories, in order.
        """
        return iter(self._factories)

    def _make_chain_for_func(self, func):
        def chain(*args, **kwargs):
            for factory in self.factories:
                converter = func(factory)(*args, **kwargs)
                if converter is not None:
                    return converter
        return functools.partial(chain, *self._args, **self._kwargs)

    def _make_chain_for_key(self, converter_key):
        return self._make_chain_for_func(
            self._converter_factory_registry[converter_key]
        )

    def __getitem__(self, converter_key):
        """
        Retrieves a callable that creates converters for the type
        associated to the given key.

        If the given key is a callable, it will be recursively invoked
        to retrieve the final callable. See :py:class:`keys.Map` for
        an example of such a key. These callable keys should accept a
        single argument, a :py:class:`ConverterFactoryRegistry`.
        """
        if callable(converter_key):
            return converter_key(self)
        else:
            return self._make_chain_for_key(converter_key)

    def __len__(self):
        return len(self._converter_factory_registry)

    def __iter__(self):
        return iter(self._converter_factory_registry)

    @classmethod
    def register(cls, converter_key):
        """
        Returns a decorator that can be used to register a callable for
        the given ``converter_key``.
        """
        def wrapper(func):
            cls._converter_factory_registry[converter_key] = func
            return func
        return wrapper


@ConverterFactoryRegistry.register(keys.CONVERT_TO_REQUEST_BODY)
def make_request_body_converter(factory):
    return factory.make_request_body_converter


@ConverterFactoryRegistry.register(keys.CONVERT_FROM_RESPONSE_BODY)
def make_response_body_converter(factory):
    return factory.make_response_body_converter


@ConverterFactoryRegistry.register(keys.CONVERT_TO_STRING)
def make_string_converter(factory):
    return factory.make_string_converter
